var remoteDirTree = {
  data     : new Array(),
  rowCount : 0,

  getCellText         : function(row, column)       { return this.data[row].leafName; },
  getLevel            : function(row)               { return this.data[row].level; },
  getParentIndex      : function(row)               { return this.data[row].parentIndex; },
  getImageSrc         : function(row, col)          { },
  getCellProperties   : function(row, col, props)   { },
  getColumnProperties : function(colid, col, props) { },
  getRowProperties    : function(row, props)        { },
  hasNextSibling      : function(row, nextrow)      { return this.data[row].hasNext; },
  isContainer         : function(row)               { return true; },
  isContainerEmpty    : function(row)               { return this.data[row].empty; },
  isContainerOpen     : function(row)               { return this.data[row].open; },
  isSeparator         : function(row)               { return false; },
  isSorted            : function(row)               { return false; },
  setTree             : function(treebox)           { this.treebox = treebox; },
  canDrop             : function(index, orient)     { return false; },

  toggleOpenState     : function(row)               {
    if (!gFtp.isConnected || gFtp.isListing()) {
      return;
    }

    if (this.isContainerOpen(row)) {
      var level     = this.data[row].level;
      var lastChild = row;

      while (lastChild + 1 < this.rowCount && this.data[lastChild + 1].level > level) {            // find last index in same level as collapsed dir
        ++lastChild;
      }

      this.data[row].children = this.data.splice(row + 1, lastChild - row);                        // get rid of subdirectories from view
      this.updateParentIndices();
      this.rowCount = this.data.length;
      this.treebox.rowCountChanged(row, -(lastChild - row));

      this.data[row].open = false;
      this.treebox.invalidateRow(row);                                                             // update row

      if (gRemotePath.value.indexOf(this.data[row].path) == 0 && gRemotePath.value != this.data[row].path
       && gRemotePath.value.match(/\x2f/g).length > this.data[row].level) {
        gRemotePath.value = this.data[row].path;                                                   // we were in a subdirectory and we collapsed
        this.selection.select(row);
        this.treebox.ensureRowIsVisible(row);
        remoteTree.updateView();
      } else if (gRemotePath.value == this.data[row].path) {
        this.selection.select(row);
        this.treebox.ensureRowIsVisible(row);
      }
    } else {
      if (this.data[row].children) {                                                               // stored from before
        for (var x = this.data[row].children.length - 1; x >= 0; --x) {
          this.data.splice(row + 1, 0, this.data[row].children[x]);
        }

        this.updateParentIndices();
        this.rowCount           = this.data.length;
        this.treebox.rowCountChanged(row + 1, this.data[row].children.length);
        this.data[row].children = null;
        this.data[row].open     = true;
        this.treebox.invalidateRow(row);
      } else {
        gFtp.list(this.data[row].path, "remoteDirTree.toggleOpenState2(" + row + ")", false);      // get data for this directory
      }
    }
  },

  toggleOpenState2 : function(row) {
    var newDirectories = new Array();

    for (var x = 0; x < gFtp.listData.length; ++x) {                                               // pick out the directories
      if (gFtp.listData[x].isDirectory()) {
        newDirectories.push(gFtp.listData[x]);
      }
    }

    if (newDirectories.length == 0)  {                                                             // no subdirectories
      this.data[row].empty = true;
      this.data[row].open  = false;
    } else {                                                                                       // has subdirectories
      for (var x = 0; x < newDirectories.length; ++x) {
        var path          = gFtp.constructPath(this.data[row].path, newDirectories[x].leafName);
        newDirectories[x] = { open        : false,
                              empty       : false,
                              hasNext     : true,
                              parentIndex : -1,
                              children    : null,
                              path        : path,
                              leafName    : newDirectories[x].leafName,
                              parent      : this.data[row].path,
                              level       : path.match(/\x2f/g).length,
                              sortPath    : path.replace(/\x2f/g, "\x01").toLowerCase() };
      }

      newDirectories.sort(directorySort);
      newDirectories[newDirectories.length - 1].hasNext = false;                                   // last one doesn't have a next

      for (var x = newDirectories.length - 1; x >= 0; --x) {
        this.data.splice(row + 1, 0, newDirectories[x]);
      }

      this.updateParentIndices();
      this.rowCount       = this.data.length;
      this.treebox.rowCountChanged(row + 1, newDirectories.length);
      this.data[row].open = true;
    }

    this.treebox.invalidateRow(row);

    if (this.updateViewAfter) {
      this.updateViewAfter = false;
      remoteTree.updateView2();
    }
  },

  updateParentIndices : function () {
    for (var x = 0; x < this.data.length; ++x) {
      this.data[x].parentIndex = this.indexOfPath(this.data[x].parent);                            // ah, beautiful
    }
  },

  indexOfPath : function(path) {                                                                   // binary search to find a path in the remoteDirTree
    var left  = 0;
    var right = this.data.length - 1;
    path      = path.replace(/\x2f/g, "\x01").toLowerCase();

    while (left <= right) {
      var mid      = Math.floor((left + right) / 2);
      var dataPath = this.data[mid].sortPath;
      if (dataPath == path || dataPath + "/" == path || dataPath == path + "/") {
        return mid;
      } else if (path < dataPath) {
        right = mid - 1;
      } else if (path > dataPath) {
        left  = mid + 1;
      }
    }

    return -1;
  },

  cdup : function() {
    var parentIndex = this.getParentIndex(this.selection.currentIndex);

    if (parentIndex != -1) {
      this.changeDir(this.data[parentIndex].path);
      this.selection.select(parentIndex);
    }
  },

  updateViewAfter : false,

  changeDir : function (path) {
    if (!gFtp.isConnected || gFtp.isListing()) {
      return;
    }

    gRemotePath.value = path;

    if (this.data.length == 0) {                                                                   // if dirTree is empty
      var oldRowCount = this.rowCount;
      this.data       = new Array();
      this.rowCount   = 0;
      this.treebox.rowCountChanged(0, -oldRowCount);

      this.data.push({ open        : false,                                                        // restart the tree
                       empty       : false,
                       hasNext     : false,
                       parentIndex : -1,
                       children    : null,
                       path        : "/",
                       leafName    : "/",
                       parent      : "",
                       level       : 0,
                       sortPath    : "/".replace(/\x2f/g, "\x01").toLowerCase() });

      this.rowCount = 1;
      this.treebox.rowCountChanged(0, 1);
      this.selection.select(0);
    }
                                                                                                   // error checking here for correct values in path
    gRemotePath.value = gRemotePath.value.replace(/\x5c/g, "/");                                   // shouldn't have backslashes

    if (gRemotePath.value != '/' && gRemotePath.value.charAt(gRemotePath.value.length - 1) == '/') {
      gRemotePath.value = gRemotePath.value.substring(0, gRemotePath.value.length - 1);            // cannot end with '\' ruins levels in dir tree
    }

    if (gRemotePath.value.charAt(0) != '/') {                                                      // has to start with '/'
      gRemotePath.value = '/' + gRemotePath.value;
    }

    for (var x = 0; x < this.data.length; ++x) {                                                   // open parent directories til we find the directory
      for (var y = this.data.length - 1; y >= x; --y) {
        if (gRemotePath.value.indexOf(this.data[y].path) == 0) {
          x = y;
          break;
        }
      }

      if (gRemotePath.value.indexOf(this.data[x].path) == 0) {
        if (!this.data[x].open && !this.data[x].empty && gRemotePath.value != this.data[x].path) {
          gRemotePath.value = this.data[x].path;
          this.selection.select(x);
          gFtp.list(this.data[x].path, "remoteDirTree.toggleOpenState2(" + x + ");remoteDirTree.changeDir('" + path + "')", false);
          return;
        } else if (gRemotePath.value == this.data[x].path) {
          gRemotePathFocus = gRemotePath.value;                                                    // directory approved
          gFormHistory.addEntry(gRemotePath.getAttribute("autocompletesearchparam"), gRemotePath.value);

          if (!this.data[x].open && !this.data[x].empty) {
            this.updateViewAfter = true;
            this.toggleOpenState(x);
          } else {
            remoteTree.updateView();
          }

          return;
        }
      }
    }

    this.dontPanic();
  },

  dontPanic : function () {                                                                        // we haven't found the directory in the conventional
    gRemotePathFocus = gRemotePath.value;                                                          // sense: don't freak out, just create directories
    this.treebox.rowCountChanged(0, -this.rowCount);                                               // above as if it were there and list
    this.rowCount    = 0;
    this.data        = new Array();
    this.data.push({ open        : true,                                                           // restart the tree
                     empty       : false,
                     hasNext     : false,
                     parentIndex : -1,
                     children    : null,
                     path        : "/",
                     leafName    : "/",
                     parent      : "",
                     level       : 0,
                     sortPath    : "/".replace(/\x2f/g, "\x01").toLowerCase() });

    var paths     = gRemotePath.value.split("/");
    var parentDir = "/";

    for (var x = 0; x < paths.length; ++x) {
      if (paths[x] != "") {
        var path = gFtp.constructPath(parentDir, paths[x]);
        this.data.push({ open        : true,
                         empty       : false,
                         hasNext     : false,
                         parentIndex : -1,
                         children    : null,
                         path        : path,
                         leafName    : paths[x],
                         parent      : parentDir,
                         level       : path.match(/\x2f/g).length,
                         sortPath    : path.replace(/\x2f/g, "\x01").toLowerCase() });

        parentDir += (parentDir != "/" ? "/" : "") + paths[x];
      }
    }

    this.data[this.data.length - 1].open = false;                                                  // make last directory closed

    this.updateParentIndices();
    this.rowCount = this.data.length;                                                              // update tree
    this.treebox.rowCountChanged(0, this.rowCount);
    this.selection.select(this.rowCount - 1);

    this.updateViewAfter = true;                                                                   // open up the last directory
    this.toggleOpenState(this.data.length - 1);
  },

  // ************************************************* events *****************************************************

  click : function(event) {
    var row = {};    var col = {};    var child = {};
    this.treebox.getCellAt(event.pageX, event.pageY, row, col, child);
    var index = this.selection.currentIndex;

    if (index == row.value && this.data[index].path != gRemotePath.value) {                        // index == row.value in case were collapsing the folder
      this.changeDir(this.data[index].path);
    }
  },

  remoteDirKeyDownIndex  : -1,
  remoteDirKeyDownOpen   : -1,

  keyDown  : function(event) {
    if (!gFtp.isConnected) {
      return;
    }

    this.remoteDirKeyDownIndex = this.selection.currentIndex;
    this.remoteDirKeyDownOpen  = this.data[this.remoteDirKeyDownIndex].open;
  },

  keyPress : function(event) {
    if (!gFtp.isConnected) {
      return;
    }

    var index = this.selection.currentIndex;

    switch (event.keyCode) {
      case 37:                                                                                     // left arrow
        if (!this.data[index].open && this.getParentIndex(index) != -1) {
          gRemotePath.value = this.data[this.getParentIndex(index)].path;
          remoteTree.updateView();
          this.selection.select(index);
        }
        break;
      case 38:                                                                                     // up arrow
      case 40:                                                                                     // down arrow
        if (event.keyCode == 38 && !index) {
          return;
        }

        if (event.keyCode == 40 && index == this.data.length - 1) {
          this.selection.select(index - 1);                                                        // this shouldn't work but it does
          return;
        }

        var x = event.keyCode == 38 ? -1 : 1;

        gRemotePath.value = this.data[index + x].path;
        remoteTree.updateView();
        this.selection.select(index);
        break;
      case 39:                                                                                     // right arrow
        if (this.remoteDirKeyDownOpen && !this.data[index].empty) {
          gRemotePath.value = this.data[index + 1].path;
          remoteTree.updateView();
        }
        break;
      case 8:                                                                                      // backspace
        this.cdup();
        break;
      case 116:                                                                                    // F5
        event.preventDefault();
        remoteTree.refresh();
        break;
    }
  }
};
